@@ -26,7 +26,7 @@ class ComponentRun(Base):
         inputs: typing.List[IOPointer] = [],
         outputs: typing.List[IOPointer] = [],
         git_hash: str = None,
-        git_tags: typing.List[str] = [],
+        git_tags: typing.List[str] = None,
         code_snapshot: str = None,
         id: str = None,
         stale: typing.List[str] = [],
@@ -239,4 +239,8 @@ class ComponentRun(Base):
             params["end_timestamp"] = params["end_timestamp"].strftime(
                 "%Y-%m-%d %l:%M:%S%z"
             )
+        for inp in params["inputs"]:
+            del inp["value"]
+        for out in params["outputs"]:
+            del out["value"]
         return json.dumps(params)
