@@ -2,17 +2,19 @@ from mltrace.db.models import PointerTypeEnum
 from mltrace.entities.base import Base
 
 import json
-import pprint
+import typing
 
 
 class IOPointer(Base):
     def __init__(
         self,
         name: str,
+        value: typing.Any = "",
         pointer_type: PointerTypeEnum = PointerTypeEnum.UNKNOWN,
         flag: bool = False,
     ):
         self._name = name
+        self._value = value
         self._pointer_type = pointer_type
         self._flag = flag
 
@@ -20,6 +22,10 @@ class IOPointer(Base):
     def name(self) -> str:
         return self._name
 
+    @property
+    def value(self) -> typing.Any:
+        return self._value
+
     @property
     def pointer_type(self) -> PointerTypeEnum:
         return self._pointer_type
@@ -30,5 +36,6 @@ class IOPointer(Base):
 
     def __repr__(self):
         params = self.to_dictionary()
+        del params["value"]
         params["pointer_type"] = params["pointer_type"].value
         return json.dumps(params)
